;; -*- lexical-binding: t; -*-
;;
;; SPDX-License-Identifier: GPL-3.0-or-later


;;;; Helper Functions

(defun property-browser-text-and-overlay-property-symbols-at (pos)
  "Return a list of properties at POS in the current buffer.

The list contains just the the property symbols not their
assigned values."

  (let* ((text-props    (text-properties-at pos))
	 (overlay-props (mapcan #'overlay-properties (overlays-at pos)))
	 (all-props     (nconc text-props overlay-props))
	 (prop-symbols  (mapcar #'car (seq-partition all-props 2))))
  (delete-dups prop-symbols)))


(defun property-browser-get-char-properties-and-overlays (pos)
  "Return the char properties in the current buffer at POS.

The return value is an alist with elements (property . (value
. source)), where source is either nil, if the value was taken
from a text property, or an overlay."

  (let* ((get-prop     (apply-partially #'get-char-property-and-overlay pos))
	 (prop-symbols (property-browser-text-and-overlay-property-symbols-at
		        pos))
	 (prop-values  (mapcar get-prop prop-symbols)))
    (seq-mapn #'cons prop-symbols prop-values)))


(defun property-browser-generate-new-at-point-buffer ()
  "Generate a new buffer for ‘list-properties-at-point-minor-mode’.

Return the newly generated buffer."

  (let* ((at-point-buffer-name (concat "*properties-at-point-"
				      (buffer-name) "*"))
	 (at-point-buffer (generate-new-buffer at-point-buffer-name))
	 (watched-buffer (current-buffer)))

    (with-current-buffer watched-buffer
      (setq-local property-browser-at-point-buffer at-point-buffer))

    (with-current-buffer at-point-buffer
      (setq-local property-browser-watched-buffer watched-buffer)
      (tabulated-list-mode)
      (setq tabulated-list-format [("O" 3 t) ; overlay-marker
				   ("Name" 25 t)
				   ("Value" 50 t)])
      (tabulated-list-init-header))

    at-point-buffer))


(defun property-browser-get-at-point-buffer-create ()
  "Return for the current buffer the buffer to show properties in.

See the ‘list-properties-at-point-minor-mode’."

  (let ((at-point-buffer property-browser-at-point-buffer))
    (unless (buffer-live-p at-point-buffer)
      (setq at-point-buffer (property-browser-generate-new-at-point-buffer)))
    at-point-buffer))


(defun property-browser-update-at-point-buffer (at-point-buffer rows)
  "Replace existing data in AT-POINT-BUFFER with ROWS.

ROWS is a list of elements as expected by ‘tabulated-list-mode’
with the format used by ‘list-properties-at-point-minor-mode’:

  ((id-or-nil [overlay-marker, name, value])…)"

    (with-current-buffer at-point-buffer
      (setq tabulated-list-entries rows)
      (tabulated-list-print)))


(defun property-browser-make-rows-from-properties-at (pos)
  "Return rows for use by ‘list-properties-at-point-minor-mode’.

POS referes to a buffer position in the current buffer.

Each row has the format expected by ‘tabulated-list-mode’ as used
by ‘list-properties-at-point-minor-mode’:

  (id-or-nil [overlay-marker, name, value])"

  (let ((make-row (lambda (prop)
		    (let* ((name    (symbol-name (car  prop)))
			   (value   (format "%s" (cadr prop)))
			   (overlay (cddr prop))
			   (overlay-marker (if overlay "x" "")))
		      (list nil
			    (vector overlay-marker name value)))))
	(props (property-browser-get-char-properties-and-overlays pos)))
    (mapcar make-row props)))


(defun property-browser-post-command-hook-handler ()
 "Update the buffer showing the properties at point."

  (let ((at-point-buffer (property-browser-get-at-point-buffer-create))
	(rows (property-browser-make-rows-from-properties-at (point))))
    (property-browser-update-at-point-buffer at-point-buffer rows)))


(defun property-browser-get-property-symbols-in-range (start end)
  "Return properties set in the current buffer between START and END.

The returned list contains just the property symbols not the
values."

  (let* ((get-prop-symbols
	  #'property-browser-text-and-overlay-property-symbols-at)
	 (points (number-sequence start end))
	 (prop-symbols (mapcan get-prop-symbols points)))
    (delete-dups prop-symbols)))



(defun property-browser-group-by (fn s)
  "Group consecutive runs of ‘equal’ elements in S.

The function FN is called on every element of S to produce a
value that is used for the comparison of equality."

  (let ((acc-fn (lambda (acc elem)
		  (if (null acc) (list (list elem))
		    (let ((head (car acc))
			  (tail (cdr acc)))
		      (if (equal (funcall fn (car head))
				 (funcall fn elem))
			  (cons (cons elem head) tail)
			(cons (list elem)
			      (cons (seq-reverse head)
				    tail))))))))
    (let* ((groups (seq-reduce acc-fn s nil))
	   ;; The last group added at the head will still be in
	   ;; reverse order after ‘seq-reduce’ finished.
	   (head   (seq-reverse (car groups)))
	   (tail   (cdr groups)))
      (seq-reverse (cons head tail)))))


(defun property-browser-single-property-values-in-range (prop start end)
  "Return all values of PROP between START and END in the current buffer.

The result is a list of alists.  Each alist has the form:

  `((start   . ,start)
    (end     . ,end)
    (value   . ,value)
    (overlay . ,overlay))

The overlay value is either nil for a text property or the
overlay from which the value was taken from."

  (let* ((get-prop-and-overlay
	  (lambda (pos)
	    (cons pos (get-char-property-and-overlay pos prop))))
	 (get-pos     #'car)
	 (get-value   #'cadr)
	 (get-overlay #'cddr)
	 (value-and-overlay #'cdr)

	 (range   (number-sequence start end))
	 (props   (mapcar get-prop-and-overlay range))
	 (grouped (property-browser-group-by value-and-overlay props))

	 (condense-group-extracting-start-and-end
	  (lambda (g)
	    (let* ((first   (first g))
		   (last    (car (last g)))

		   (start   (funcall get-pos     first))
		   (end     (funcall get-pos     last))
		   (value   (funcall get-value   first))
		   (overlay (funcall get-overlay first)))

	      (list (cons 'start   start)
		    (cons 'end     end)
		    (cons 'value   value)
		    (cons 'overlay overlay))))))

    (mapcar condense-group-extracting-start-and-end grouped)))


(defun property-browser-property-values-in-range (start end)
  "Return properties between START and END in the current buffer.

The return value is an alist of the form:

  (('prop . ('start   . start-pos)
            ('end     . end-pos)
            ('value   . value)
            ('overlay . overlay)))

Duplicated “prop” keys are possible.  In this case “value” or
“overlay” may be different, or the property does not extend over
the range in one continues block."

  (let ((get-props
	 (lambda (prop)
	   (cons prop
		 (property-browser-single-property-values-in-range prop
								   start
								   end))))
	(prop-symbols (property-browser-get-property-names-in-range start
								    end)))
    (mapcar get-props prop-symbols)))


(defun property-browser-compare-numerical (a b)
  "Compare two ‘tabulated-list-mode’ elements numerical.

The comparison is done honouring the current
‘tabulated-list-mode’ sort column."

  (let* ((sort-column (car tabulated-list-sort-key))
	 (n (tabulated-list--column-number sort-column)))
    (< (string-to-number (aref (cadr a) n))
       (string-to-number (aref (cadr b) n)))))


(defun property-browser-compare-string (a b)
  "Compare two ‘tabulated-list-mode’ elements as strings.

The comparison is done honouring the current
‘tabulated-list-mode’ sort column."

  ;; FIXME: Isnʼt this the default and, therefore, redundant?
  (let* ((sort-column (car tabulated-list-sort-key))
	 (n (tabulated-list--column-number sort-column)))
    (string< (aref (cadr a) n)
	     (aref (cadr b) n))))


(defun property-browser-get-list-in-range-buffer-create ()
  "Return for the current buffer the buffer to show properties in.

See the ‘list-properties-in-range’ function."

  (let ((buffer (get-buffer-create (concat "*properties-in-range-"
					   (buffer-name) "*"))))
    (with-current-buffer buffer
      (let* ((table-format [("Name"  15 t)
			    ("Value" 40 property-browser-compare-string)
			    ("Start"  6 property-browser-compare-numerical)
			    ("End"    6 property-browser-compare-numerical)
			    ("O" 3)])) ; overlay-marker
	(tabulated-list-mode)
	(setq tabulated-list-format table-format)
	(tabulated-list-init-header)))
    buffer))


;;;; Interactive Commands

(defun property-browser-display-at-point-buffer ()
  "Display the properties-at-point buffer for the current buffer."
  (interactive)

  (let ((at-point-buffer property-browser-at-point-buffer))
    (if (buffer-live-p at-point-buffer) (display-buffer at-point-buffer)
      (user-error "Property browser buffer dead or non-existant."))))


(defun list-properties-in-range (start end)
  "List properties in the current region."
  (interactive "r")

  (let* ((make-row (lambda (name d)
		     (let* ((start   (format "%s" (cdr (assq 'start d))))
			    (end     (format "%s" (cdr (assq 'end   d))))
			    (value   (format "%s" (cdr (assq 'value d))))
			    (overlay (cdr (assq 'overlay d)))
			    (overlay-mark (if (overlayp overlay) "x" "")))
		       (list nil
			     (vector name value start end overlay-mark)))))

	 (props (property-browser-property-values-in-range start end))
	 (rows (cl-loop for (prop-symbol . data) in props
			for prop-name = (symbol-name prop-symbol)
			nconc
			(cl-loop for d in data
				 collect (funcall make-row prop-name d))))
	 (buffer (property-browser-get-list-in-range-buffer-create)))

    (with-current-buffer buffer
      (setq tabulated-list-entries rows)
      (tabulated-list-print))
    (display-buffer buffer)))


(define-minor-mode list-properties-at-point-minor-mode
  "List properties at point."
  :lighter " pbrowse"
  (cond
   (list-properties-at-point-minor-mode
    (property-browser-generate-new-at-point-buffer)
    (property-browser-display-at-point-buffer)

    (add-hook 'post-command-hook
	      #'property-browser-post-command-hook-handler
	      nil :local))
   (t
    (when (buffer-live-p property-browser-at-point-buffer)
      (kill-buffer property-browser-at-point-buffer))
    (setq-local property-browser-at-point-buffer nil)

    (remove-hook 'post-command-hook
		 #'property-browser-post-command-hook-handler
		 :local))))


;;;; Footer

(provide 'property-browser)

;;; property-browser.el ends here
